// SDIunrooted.java
// Speciation versus Duplication Inference for a possibly unrooted tree
//
// Copyright (C) 1999-2001 Washington University School of Medicine
// and Howard Hughes Medical Institute
// All rights reserved
//
// Created: 2000
// Author: Christian M. Zmasek
// zmasek@genetics.wustl.edu
// http://www.genetics.wustl.edu/eddy/people/zmasek/



// Requires JDK 1.2 or greater.


package forester.tools;


import forester.tree.*;
import forester.atv.*;

import java.io.*;
import java.util.Vector;



/**

Allows to infer duplications - speciations on a unrooted gene tree.
It reroots the gene trees on each of its branches and performs SDIse
or OE on each of the resulting trees. Trees which minimize a certain
criterion are returned as the "correctly" rooted ones. The criterions are:
<ul>
<li> Sum of duplications
<li> Mapping cost L
<li> Tree height - which is the largest distance from root to external node
     (minimizing of which is the same as "midpoint rooting")
</ul>

@see SDIse
@see OE
@see SDI

@author Christian M. Zmasek

@version 1.200 -- last modified: 02/02/01

*/
public class SDIunrooted {

    // How many resulting trees "main" should return/display.
    private final static int     TREES_TO_RETURN = 5;

    private final static boolean TIME            = true,
                                 VERBOSE         = false;

    private final static double  ZERO_DIFF       = 1.0E-10; // Due to inaccurate
                                                            // calculations on
                                                            // Java's side, not
                                                            // everything that should
                                                            // be 0.0 is 0.0.

    private int    count_,
                   min_dup_,
                   min_cost_;
    private double min_height_,
                   min_diff_;
    private long   time_sdi;



    /**

    Default contructor which creates an "empty" SDIunrooted object.

    */
    public SDIunrooted() {
        count_      = -99;
        min_dup_    = -99;
        min_cost_   = -99;
        min_height_ = -99.0;
        min_diff_   = +99.0;
        time_sdi    = 0;
    }



    /**

    Returns the number of differently rooted trees which minimize
    the (rooting) "criterion" - as determined by method "infer".

    @see #infer(Tree,Tree,boolean,boolean,boolean,boolean,int,boolean)

    @return number of differently rooted trees which minimized
            the criterion

    */
    public int getCount() {
        return count_;
    }



    /**

    Returns the minimal number of duplications - as determined
    by method "infer".
    <p>
    <B>IMPORTANT</B>: If the tree is not rooted by minimizing
    the sum of duplications or the mapping cost L, then this
    number is NOT NECESSARILY the MINIMAL number of duplications.

    @see #infer(Tree,Tree,boolean,boolean,boolean,boolean,int,boolean)

    @return (minimal) number of duplications

    */
    public int getMinimalDuplications() {
        return min_dup_;
    }



    /**

    Returns the minimal mapping cost L - as determined
    by method "infer" - if minimize_mapping_cost is set to true.
    <p>
    (Last modified: 11/07/00)

    @see #infer(Tree,Tree,boolean,boolean,boolean,boolean,int,boolean)

    @return the minimal mapping cost "L" -- IF calculated by "infer"

    */
    public int getMinimalMappingCost() {
        return min_cost_;
    }



    /**

    Returns the minimal tree height - as determined
    by method "infer" - if minimize_height is set to true.
    <B>IMPORTANT</B>: If minimize_mapping_cost or minimize_sum_of_dup
    are also set to true, then this returns the minimal tree height
    of the trees which minimize the first criterion.
    <p>
    (Last modified: 01/12/00)

    @see #infer(Tree,Tree,boolean,boolean,boolean,boolean,int,boolean)

    @return the minimal tree height -- IF calculated by "infer"

    */
    public double getMinimalTreeHeight() {
        return min_height_;
    }



    /**

    Returns the (absolue value of the) minimal difference in tree heights
    of the two subtrees at the root (of the (re)rooted gene tree)
    - as determined by method "infer" - if minimize_height is set to true.
    <p>
    If a tree is midpoint rooted this number is zero.
    <p>
    <B>IMPORTANT</B>: If minimize_mapping_cost or minimize_sum_of_dup
    are also set to true, then this returns the minimal difference
    in tree heights of the trees which minimize the first criterion, and
    is therefore not necessarily zero.
    <p>
    (Last modified: 01/22/00)

    @see #infer(Tree,Tree,boolean,boolean,boolean,boolean,int,boolean)

    @return the minimal difference in tree heights -- IF calculated by "infer"

    */
    public double getMinimalDiffInSubTreeHeights() {
        return min_diff_;
    }



    /**

    Returns the sum of times (in ms) needed to run method infer of class SDI.
    Final variable TIME needs to be set to true.

    @return sum of times (in ms) needed to run method infer of class SDI

    */
    public long getTimeSumSDI() {
        return time_sdi;
    }



    /**

    Infers gene duplications on a possibly unrooted gene Tree gene_tree.
    The tree is rooted be minimizing either the sum of duplications, the
    mapping cost L, or the tree height (or combinations thereof).
    If return_trees is set to true, it returns an array of possibly
    more than one differently rooted Trees.
    <br>
    The maximal number of returned trees is set with max_trees_to_return.
    <br>
    If use_eulenstein is set to true, Eulenstein's algorithm is used,
    otherwise our SDI algorithm for duplication inference is used.
    <br>
    Tree species_tree is a species Tree to which the gene
    Tree gene_tree is compared to.
    <br>
    If both minimize_sum_of_dup and minimize_mapping_cost are true,
    the tree is rooted by minimizing the mapping cost L.
    <br>
    If minimize_sum_of_dup, minimize_mapping_cost, and minimize_height
    are false tree gene_tree is assumed to be alreadty rooted and no
    attempts at rooting are made, and only one tree is returned.
    <br>
    <p>
    Conditions:
    </p>
    <ul>
    <li> Both Trees must be completely binary (except deepest node of gene tree)
    <li> The species Tree must be rooted
    <li> Both Trees must have species names in the species name fields of their nodes
    </ul>

    (Last modified: 01/25/01)

    @param gene_tree             a binary (except deepest node)
                                 gene Tree
    @param species_tree          a rooted binary species Tree
    @param minimize_mapping_cost set to true to root by minimizing the
                                 mapping cost L (and also the
                                 sum of duplications)
    @param minimize_sum_of_dup   set to true to root by minimizing
                                 the sum of duplications
    @param minimize_height       set to true to root by minimizing the
                                 tree height - if minimize_mapping_cost is
                                 set to true or minimize_sum_of_dup is set
                                 to true, then out of the resulting trees with
                                 minimal mapping cost or minimal number of
                                 duplications the tree with the minimal height
                                 is chosen
    @param return_trees          set to true to return Array of Trees,
                                 otherwise null is returned
    @param max_trees_to_return   maximal number of Trees to return
                                 (=maximal size of returned Array)
                                 must be no lower than 1
    @param use_eulenstein        set to true to use Eulenstein's algorithm
                                 instead of our SDIse

    @return array of rooted Trees with duplication vs. speciation assigned if
            return_trees is set to true, null otherwise

    */
    public Tree[] infer( Tree gene_tree,
                         Tree species_tree,
                         boolean minimize_mapping_cost,
                         boolean minimize_sum_of_dup,
                         boolean minimize_height,
                         boolean return_trees,
                         int max_trees_to_return,
                         boolean use_eulenstein ) throws Exception {


        SDI                  sdi                   = null;
        Vector               trees                 = new Vector();
        Tree[]               tree_array            = null;
        Node[]               nodes                 = null;
        PreorderTreeIterator it                    = null;
        Tree                 g                     = null,
                             species_tree_stripped = null;
        Node                 n                     = null,
                             orig_extnode0         = null;
        int                  duplications          = 0,
                             cost                  = 0,
                             counter               = 0,
                             min_duplications      = Integer.MAX_VALUE,
                             min_cost              = Integer.MAX_VALUE,
                             nodecount0            = Node.getNodeCount(),
                             j                     = 0,
                             x                     = 0;
        double               height                = 0.0,
                             diff                  = 0.0,
                             min_height            = Double.MAX_VALUE,
                             min_diff              = 0.0;
        double[]             height_n_diff         = new double[ 2 ];
        long                 time0                 = 0;
        boolean              no_need_to_root       = false,
                             smaller               = false,
                             equal                 = false;


        if ( max_trees_to_return < 1 ) {
            max_trees_to_return = 1;
        }

        if ( minimize_mapping_cost && minimize_sum_of_dup ) {
            minimize_sum_of_dup = false;
        }

        if ( !minimize_mapping_cost && !minimize_sum_of_dup
        && !minimize_height ) {
            max_trees_to_return = 1;
            no_need_to_root     = true;
        }


        g = gene_tree.copyTree();

        if ( g.getNumberOfExtNodes() <= 1 ) {
            g.setRooted( true );
            tree_array = new Tree[ 1 ];
            tree_array[ 0 ] = g;
            return tree_array;
        }

        species_tree_stripped = species_tree.copyTree();

        // Removes from species_tree all species not found in g.
        SDI.stripTree( g, species_tree_stripped );

        if ( !no_need_to_root ) {
            orig_extnode0 = g.getExtNode0();
            g.reRoot( orig_extnode0 );
            try {
                it = new PreorderTreeIterator( g );
            }
            catch ( Exception e )  {
                System.err.println( "Could not create iterator. Terminating." );
                System.exit( -1 );
            }

            nodes = new Node[ ( 2 * g.getNumberOfExtNodes() ) - 2 ];
            j = 0;
            while ( !it.isDone() ) {
                n = it.currentNode();
                if ( !n.isPseudoNode() && !n.isRoot() && n != orig_extnode0 ) {
                    nodes[ j++ ] = n;
                }
                it.next();
            }
            x = j;
        }
        else {
            x = 1;
        }

        n = null;

        F: for ( j = 0; j < x; ++j ) {

            if ( VERBOSE ) {
                System.out.print( j + ":" );
            }

            if ( !no_need_to_root ) {
                n = nodes[ j ];
                g.reRoot( orig_extnode0 );
                g.reRoot( n );


                // This ensures that the root is in the middle,
                // which is a condition for method moveRoot.
                if ( minimize_height
                && ( g.getRoot().getChild1().getDistanceToParent()
                != g.getRoot().getChild2().getDistanceToParent() ) ) {
                    Node   root   = g.getRoot();
                    Node   child1 = root.getChild1(),
                           child2 = root.getChild2();
                    double d1     = child1.getDistanceToParent(),
                           d2     = child2.getDistanceToParent();
                    if ( d1 < 0.0 ) {
                        d1 = 0.0;
                    }
                    if ( d2 < 0.0 ) {
                        d2 = 0.0;
                    }
                    child1.setDistanceToParent( ( d1 + d2 ) / 2.0 );
                    child2.setDistanceToParent( ( d1 + d2 ) / 2.0 );
                }
            }

            if ( use_eulenstein ) {
                sdi = new OE( g, species_tree_stripped );
            }
            else {
                sdi = new SDIse( g, species_tree_stripped );
            }

            if ( minimize_mapping_cost || minimize_sum_of_dup
            || no_need_to_root ) {
                if ( TIME ) {
                    time0 = System.currentTimeMillis();
                }

                duplications = sdi.infer( false ); // Stripping has already been done.

                if ( TIME ) {
                    time_sdi += ( System.currentTimeMillis() - time0 );
                }
                if ( VERBOSE ) {
                    System.out.print( " d=" + duplications );
                }
            }


            if ( minimize_mapping_cost ) {

                cost = sdi.computeMappingCost();
                if ( VERBOSE ) {
                    System.out.print( " L=" + cost );
                }

                if ( minimize_height
                && ( cost == min_cost || cost < min_cost ) ) {

                    diff = g.calculateRealHeight();
                    height_n_diff = moveRoot( g, diff, g.getRealHeight() );
                    height = height_n_diff[ 0 ];
                    diff   = height_n_diff[ 1 ];

                    if ( VERBOSE ) {
                        System.out.print( " h=" + height );
                        System.out.print( " d=" + diff );
                    }

                }

                if ( cost == min_cost ) {
                    if ( minimize_height ) {
                        smaller = equal = false;
                        if ( height < min_height ) {
                            min_height = height;
                            counter = 1;
                            smaller = true;
                        }
                        else if ( height == min_height ) {
                            counter++;
                            equal = true;
                        }
                        if ( Math.abs( diff ) < min_diff ) {
                            min_diff = Math.abs( diff );
                        }
                    }
                    if ( return_trees ) {
                        if ( minimize_height ) {
                            if ( smaller ) {
                                trees.removeAllElements();
                                adjust( g );
                                trees.addElement( g.copyTree() );
                            }
                            else if ( equal && trees.size() < max_trees_to_return ) {
                                adjust( g );
                                trees.addElement( g.copyTree() );
                            }
                        }
                        else {
                            counter++;
                            if ( trees.size() < max_trees_to_return ) {
                                adjust( g );
                                trees.addElement( g.copyTree() );
                            }
                        }
                    }
                    else if ( !minimize_height ) {
                        counter++;
                    }
                }
                else if ( cost < min_cost ) {
                    if ( minimize_height ) {
                        min_height = height;
                        min_diff   = Math.abs( diff );
                    }
                    if ( return_trees ) {
                        adjust( g );
                        trees.removeAllElements();
                        trees.addElement( g.copyTree() );
                    }
                    counter = 1;
                    min_cost = cost;
                }
                if ( duplications < min_duplications ) {
                    min_duplications = duplications;
                }
            }


            else if ( minimize_sum_of_dup ) {

                if ( minimize_height
                && ( duplications == min_duplications
                || duplications < min_duplications ) ) {

                    diff = g.calculateRealHeight();
                    height_n_diff = moveRoot( g, diff, g.getRealHeight() );
                    height = height_n_diff[ 0 ];
                    diff   = height_n_diff[ 1 ];

                    if ( VERBOSE ) {
                        System.out.print( " h=" + height );
                        System.out.print( " d=" + diff );
                    }

                }

                if ( duplications == min_duplications ) {
                    if ( minimize_height ) {
                        smaller = equal = false;
                        if ( height < min_height ) {
                            min_height = height;
                            counter = 1;
                            smaller = true;
                        }
                        else if ( height == min_height ) {
                            counter++;
                            equal = true;
                        }
                        if ( Math.abs( diff ) < min_diff ) {
                            min_diff = Math.abs( diff );
                        }
                    }
                    if ( return_trees ) {
                        if ( minimize_height ) {
                            if ( smaller ) {
                                trees.removeAllElements();
                                adjust( g );
                                trees.addElement( g.copyTree() );
                            }
                            else if ( equal && trees.size() < max_trees_to_return ) {
                                adjust( g );
                                trees.addElement( g.copyTree() );
                            }
                        }
                        else {
                            counter++;
                            if ( trees.size() < max_trees_to_return ) {
                                adjust( g );
                                trees.addElement( g.copyTree() );
                            }
                        }
                    }
                    else if ( !minimize_height ) {
                        counter++;
                    }
                }
                else if ( duplications < min_duplications ) {
                    if ( minimize_height ) {
                        min_height = height;
                        min_diff   = Math.abs( diff );
                    }
                    if ( return_trees ) {
                        adjust( g );
                        trees.removeAllElements();
                        trees.addElement( g.copyTree() );
                    }
                    counter = 1;
                    min_duplications = duplications;
                }
            }


            else if ( minimize_height ) {

                diff = g.calculateRealHeight();
                height_n_diff = moveRoot( g, diff, g.getRealHeight() );
                height = height_n_diff[ 0 ];
                diff   = height_n_diff[ 1 ];

                if ( VERBOSE ) {
                    System.out.print( " h=" + height );
                    System.out.print( " d=" + diff );
                }

                if ( Math.abs( diff ) < ZERO_DIFF ) {
                     if ( TIME ) {
                         time0 = System.currentTimeMillis();
                     }
                     min_duplications = sdi.infer( false ); // Here they are obv. not _min_ dups !!!!
                     if ( TIME ) {
                         time_sdi += ( System.currentTimeMillis() - time0 );
                     }
                     if ( VERBOSE ) {
                        System.out.print( " d=" + min_duplications + "\n" );
                     }
                     min_height = height;
                     min_diff   = Math.abs( diff );
                     counter    = 1;
                     if ( return_trees ) {
                        adjust( g );
                        trees.addElement( g.copyTree() );
                     }
                     break F;
                }
            }


            else { // No rooting has been performed
                min_duplications = duplications;
                if ( return_trees ) {
                    trees.addElement( g.copyTree() );
                }
            }


            if ( VERBOSE ) {
                System.out.println( "" );
            }

        } // End of huge for loop "F".


        if ( return_trees ) {
            trees.trimToSize();
            tree_array = new Tree[ trees.size() ];
            for ( int i = 0; i < trees.size(); ++i ) {
                tree_array[ i ] = ( Tree ) trees.elementAt( i );
            }
        }

        setCount( counter );
        setMinimalDuplications( min_duplications );
        setMinimalMappingCost( min_cost );
        setMinimalTreeHeight( min_height );
        setMinimalDiffInSubTreeHeights( Math.abs( min_diff ) );

        return tree_array;

    }



    // This places to root on its branch in such a way that it
    // minimizes the tree height as good as possible.
    // PRECONDITION: the root is placed at the midpoint of its branch.
    private double[] moveRoot( Tree t, double diff, double height ) {
        Node     root           = t.getRoot();
        Node     child1         = root.getChild1(),
                 child2         = root.getChild2();
        double   d              = 0.0;
        double[] height_n_diff  = new double[ 2 ];

        if ( child1 != null && child1.getDistanceToParent() > 0.0 ) {

            d = child1.getDistanceToParent();

            if ( ( 2 * d ) > Math.abs( diff ) ) {
                child1.setDistanceToParent( d - ( diff / 2.0 ) );
                child2.setDistanceToParent( d + ( diff / 2.0 ) );
                height_n_diff[ 0 ] = height - Math.abs( diff / 2 );
                height_n_diff[ 1 ] = 0.0;
            }
            else {
                if ( diff > 0 ) {
                    child1.setDistanceToParent( 0.0 );
                    child2.setDistanceToParent( 2 * d );
                    height_n_diff[ 1 ] = diff - ( 2 * d );
                }
                else {
                    child1.setDistanceToParent( 2 * d );
                    child2.setDistanceToParent( 0.0 );
                    height_n_diff[ 1 ] = diff +  ( 2 * d );
                }
                height_n_diff[ 0 ] = height - d;
            }
        }
        else {
            height_n_diff[ 0 ] = height;
            height_n_diff[ 1 ] = diff;
        }
        return height_n_diff;
    }



    private void adjust( Tree t ) {
        t.adjustNodeCount();
        t.setExtNodes( t.getRoot().getSumExtNodes() );
    }



    private void setCount( int i ) {
        count_ = i;
    }



    private void setMinimalDuplications( int i ) {
        min_dup_ = i;
    }



    private void setMinimalMappingCost( int i ) {
        min_cost_ = i;
    }



    private void setMinimalTreeHeight( double d ) {
        min_height_ = d;
    }



    private void setMinimalDiffInSubTreeHeights( double d ) {
        min_diff_ = d;
    }



    /**

    Main method for this class.
    <p>
    (Last modified: 01/11/01)

    @param [args[0] -e to use Eulenstein's algorithm instead of SDIse]
    @param [args[0] -l to root by minimizing mapping cost L]
    @param [args[0] -d to root by minimizing sum of duplications]
    @param [args[0] -h to root by minimizing tree height]
    @param [args[0] -n input trees are in New Hampshire format instead
                       of NHX; or gene tree is in NHX, but species
                       information in gene tree is only  present in the
                       form of SWISS-PROT sequence names]
    @param args[0or1]  species tree file name (in NHX format with
                       species names in species name fields;
                       unless -n option is used)
    @param args[1or2]  gene tree file name (in NHX format with
                       species names in species name fields and sequence names
                       in sequence name fields; unless -n option is used)

    */
    public static void main( String args[] ) {



        boolean use_eulenstein      = false,
                minimize_cost       = false,
                minimize_sum_of_dup = false,
                minimize_height     = false,
                nh                  = false;

        int     r                   = 0;

        File    species_tree_file   = null,
                gene_tree_file      = null;

        Tree    gene_tree           = null,
                species_tree        = null;

        Tree[]  trees               = null;

        SDIunrooted sdiunrooted     = new SDIunrooted();

        java.text.DecimalFormat df = new java.text.DecimalFormat( "0.0#####" );
        df.setDecimalSeparatorAlwaysShown( true );


        if ( args.length < 2 || args.length > 3 ) {
            errorInCommandLine();
        }
        if ( args[ 0 ].startsWith( "-" ) ) {
            if ( args.length < 3 ) {
                errorInCommandLine();
            }
            if ( args[ 0 ].toLowerCase().indexOf( "e" ) != -1 ) {
                use_eulenstein      = true;
            }
            if ( args[ 0 ].toLowerCase().indexOf( "l" ) != -1 ) {
                minimize_cost       = true;
            }
            if ( args[ 0 ].toLowerCase().indexOf( "d" ) != -1 ) {
                minimize_sum_of_dup = true;
            }
            if ( args[ 0 ].toLowerCase().indexOf( "h" ) != -1 ) {
                minimize_height     = true;
            }
            if ( args[ 0 ].toLowerCase().indexOf( "n" ) != -1 ) {
                nh                  = true;
            }
            species_tree_file = new File( args[ 1 ] );
            gene_tree_file    = new File( args[ 2 ] );
        }
        else {
            if ( args.length > 2 ) {
                errorInCommandLine();
            }
            species_tree_file = new File( args[ 0 ] );
            gene_tree_file    = new File( args[ 1 ] );
        }

        if ( minimize_cost && minimize_sum_of_dup ) {
            minimize_sum_of_dup = false;
        }

        try {
            gene_tree = TreeHelper.readNHtree( gene_tree_file );
        }
        catch ( Exception e ) {
	        System.err.println( "Could not read " + gene_tree_file + ". Terminating." );
	        System.exit( -1 );
	    }

        try {
            species_tree = TreeHelper.readNHtree( species_tree_file );
        }
        catch ( Exception e ) {
	        System.err.println( "Could not read " + species_tree_file + ". Terminating." );
	        System.exit( -1 );
	    }

        if ( nh ) {
            TreeHelper.extractSpeciesNameFromSeqName( species_tree );
            TreeHelper.extractSpeciesNameFromSeqName( gene_tree );
        }

        TreeHelper.cleanSpeciesNamesInExtNodes( species_tree );

        if ( !minimize_cost && !minimize_sum_of_dup && !minimize_height ) {
            // In this case.... the gene tree must be rooted.
            gene_tree.setRooted( true );
        }

        // Removes from gene_tree all species not found in species_tree.
        r = SDI.stripTree( species_tree, gene_tree );
        System.out.println( "\nRemoved " + r + " external nodes from gene tree.\n" );

        try {
            trees = sdiunrooted.infer( gene_tree,
                                       species_tree,
                                       minimize_cost,
                                       minimize_sum_of_dup,
                                       minimize_height,
                                       true,
                                       TREES_TO_RETURN,
                                       use_eulenstein );
        }
        catch ( Exception e ) {
	        System.err.println( "Unexpected error during calculation of duplications." );
            System.err.println( "Stack trace: " );
            e.printStackTrace();
	        System.exit( -1 );
	    }

        System.out.println( "" );

        if ( minimize_cost ) {
            System.out.println( "Rooted by minimizing mapping cost L." );
            if ( minimize_height ) {
                System.out.println( "Selected tree(s) with minimal height out of resulting trees." );
            }
            System.out.println( "Number differently rooted trees minimizing criterion  : "
            + sdiunrooted.getCount() );
            System.out.println( "Minimal cost                                          : "
            + sdiunrooted.getMinimalMappingCost() );
            System.out.println( "Minimal duplications                                  : "
            + sdiunrooted.getMinimalDuplications() );
            if ( minimize_height ) {
                System.out.println( "Tree height                                           : "
                + df.format( sdiunrooted.getMinimalTreeHeight() ) );
                System.out.println( "Difference in subtree heights                         : "
                + df.format( sdiunrooted.getMinimalDiffInSubTreeHeights() ) );
            }

            System.out.println( "" );
        }
        else if ( minimize_sum_of_dup ) {
            System.out.println( "Rooted by minimizing sum of duplications." );
            if ( minimize_height ) {
                System.out.println( "Selected tree(s) with minimal height out of resulting trees." );
            }
            System.out.println( "Number differently rooted trees minimizing criterion        : "
            + sdiunrooted.getCount() );
            System.out.println( "Minimal duplications                                        : "
            + sdiunrooted.getMinimalDuplications() );
            if ( minimize_height ) {
                System.out.println( "Tree height                                                 : "
                + df.format( sdiunrooted.getMinimalTreeHeight() ) );
                System.out.println( "Difference in subtree heights                               : "
                + df.format( sdiunrooted.getMinimalDiffInSubTreeHeights() ) );
            }
            System.out.println( "" );
        }
        else if ( minimize_height ) {
            System.out.println( "Rooted by minimizing tree height (midpoint rooting)." );
            System.out.println( "Minimal tree height                  : "
            + df.format( sdiunrooted.getMinimalTreeHeight() ) );
            System.out.println( "Minimal difference in subtree heights: "
            + df.format( sdiunrooted.getMinimalDiffInSubTreeHeights() ) );
            System.out.println( "Duplications in midpoint rooted tree : "
            + sdiunrooted.getMinimalDuplications() );
            System.out.println( "" );
        }
        else {
            System.out.println( "No (re) rooting was performed." );
            System.out.println( "Duplications in tree: "
            + sdiunrooted.getMinimalDuplications() );
            System.out.println( "" );
        }

        ATVjframe[] atvframes = new ATVjframe[ trees.length ];

        for ( int i = 0; i < trees.length; ++i ) {
            atvframes[ i ] = new ATVjframe( trees[ i ] );
            atvframes[ i ].setTitle( "gene tree " + i );
            atvframes[ i ].showWhole();
        }

        ATVjframe atvframe_og = new ATVjframe( gene_tree );
        atvframe_og.setTitle( "original gene tree" );
        atvframe_og.showWhole();

    } // End of main.



    private static void errorInCommandLine() {
        System.out.println( "\nSDIunrooted: Error in command line.\n" );
        System.out.println( "Usage: \"SDIunrooted  [-options] <species tree file name> <gene tree file name>\"" );
        System.out.println( "\nOptions:" );
        System.out.println( " -e to use Eulenstein's algorithm instead of SDIse" );
        System.out.println( " -n input trees are in New Hampshire format instead of NHX -- or" );
        System.out.println( "    the gene tree is in NHX, but species information is" );
        System.out.println( "    only present in the form of SWISS-PROT sequence names" );
        System.out.println( " -l to root by minimizing the mapping cost L (and also the sum of duplications)" );
        System.out.println( " -d to root by minimizing the sum of duplications" );
        System.out.println( " -h to root by minimizing tree height (can be used together with -l or -d)" );
        System.out.println( "\nSpecies tree file" );
        System.out.println( " In NHX format, with species names in species name fields; unless -n option" );
        System.out.println( " is used." );
        System.out.println( "\nGene tree file" );
        System.out.println( " In NHX format, with species names in species name fields and sequence names" );
        System.out.println( " in sequence name fields; unless -n option is used." );
        System.out.println( "" );
        System.exit( -1 );
    }

}



